import json
import logging
import os
from os.path import expanduser
from typing import Any, Dict, Tuple

import boto3
from yaml import safe_load

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()

# Tags
ORBIT_PRODUCT_KEY = "Product"
ORBIT_SUBPRODUCT_KEY = "SubProduct"
ORBIT_SUBPRODUCT_EMR = "EMR"
ORBIT_SUBPRODUCT_REDSHIFT = "Redshift"
ORBIT_PRODUCT_NAME = "Orbit Workbench"
ORBIT_ENV = "Env"
AWS_ORBIT_TEAM_SPACE = "TeamSpace"


def get_properties() -> Dict[str, str]:
    """
    Returns the properties and pathnames of the current Orbit Workbench Environment.

    Parameters
    ----------
    None
        None.

    Returns
    -------
    prop : dict
        Dictionary containing pathnames for the Orbit Workbench Enviornment, Team Space, S3 Bucket Path, Orbit Workbench Source
        Repository.

    Example
    -------
    >>> from aws_orbit_sdk.common import get_properties
    >>> props = get_properties()
    """
    if "AWS_ORBIT_ENV" in os.environ.keys():
        if "AWS_ORBIT_TEAM_SPACE" not in os.environ.keys():
            logger.error("current env vars: %s", os.environ.keys())
            raise Exception("if AWS_ORBIT_ENV then AWS_ORBIT_TEAM_SPACE must be set")
        else:
            prop = dict(
                AWS_ORBIT_ENV=os.environ.get("AWS_ORBIT_ENV", ""),
                AWS_ORBIT_TEAM_SPACE=os.environ.get("AWS_ORBIT_TEAM_SPACE", ""),
            )
            if "AWS_ORBIT_S3_BUCKET" in os.environ.keys():
                prop["AWS_ORBIT_S3_BUCKET"] = os.environ.get("AWS_ORBIT_S3_BUCKET")
    else:
        # this path is used by the sagemaker notebooks where we cannot create the env variable in the context of the notebook
        home = expanduser("~")
        propFilePath = f"{home}/orbit.yaml"
        with open(propFilePath, "r") as f:
            prop = safe_load(f)["properties"]

    prop["ecs_cluster"] = f"orbit-{prop['AWS_ORBIT_ENV']}-{prop['AWS_ORBIT_TEAM_SPACE']}-cluster"
    prop["eks_cluster"] = f"orbit-{prop['AWS_ORBIT_ENV']}"
    return prop


def split_s3_path(s3_path: str) -> Tuple[str, str]:
    """
    Splits a s3 bucket path to its bucket name and its key with prefixes.

    Parameters
    ----------
    s3_path : str
        s3 bucket path to split.

    Returns
    -------
    bucket : str
        Bucket name derived from the s3 path
    key : str
        Object key and prefixes in the s3 bucket

    Example
    -------
    >>> from aws_orbit_sdk.common import split_s3_path
    >>> bucket, key = split_s3_path("s3://my-bucket/prefix/myobject.csv")
    """

    path_parts = s3_path.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)
    return bucket, key


def get_workspace() -> Dict[str, str]:
    """
    Returns workspace configuration for your given role for your Team Space in a dictionary object.

    Parameters
    ----------
    None
        None.

    Returns
    -------
    config : dict
        Dictionary object containing workspace config. on scratch-bucket, notebook_bucket, role_arn,
        instance_profile_arn, instance_profile_name, region, env_name, and team-space.

    Example
    -------
    >>> from aws_orbit_sdk.common import get_workspace
    >>> workspace = get_workspace()
    """
    ssm = boto3.client("ssm")
    props = get_properties()

    role_key = f"/orbit/{props['AWS_ORBIT_ENV']}/teams/{props['AWS_ORBIT_TEAM_SPACE']}/context"

    role_config_str = ssm.get_parameter(Name=role_key)["Parameter"]["Value"]

    config = json.loads(role_config_str)
    my_session = boto3.session.Session()
    my_region = my_session.region_name
    config["region"] = my_region
    config["env_name"] = props["AWS_ORBIT_ENV"]
    config["team_space"] = props["AWS_ORBIT_TEAM_SPACE"]

    return config


def get_scratch_database() -> str:
    """
    Creates a new scratch database for the TeamSpace located in the scratch s3 bucket.

    Parameters
    ----------
    None
        None.

    Returns
    -------
    scratch_db_name : str
        Name of the new scratch database

    Example
    -------
    >>> from aws_orbit_sdk.common import get_scratch_database
    >>> scratch_database = get_scratch_database()
    """
    glue = boto3.client("glue")
    response = glue.get_databases()
    workspace = get_workspace()
    scratch_db_name = f"scratch_db_{workspace['env_name']}_{workspace['team_space']}".lower().replace("-", "_")
    new_location = f"s3://{workspace['ScratchBucket']}/{workspace['team_space']}/{scratch_db_name}"
    for db in response["DatabaseList"]:
        if db["Name"].lower() == scratch_db_name:
            if new_location == db["LocationUri"]:
                return scratch_db_name
            else:
                ## scratch database left from previous teamspace creation , will delete it.
                glue.delete_database(Name=scratch_db_name)

    response = glue.create_database(
        DatabaseInput={
            "Name": scratch_db_name,
            "Description": f"scratch database for TeamSpace {workspace['env_name']}.{workspace['team_space']}",
            "LocationUri": new_location,
            "Parameters": {
                ORBIT_PRODUCT_KEY: ORBIT_PRODUCT_NAME,
                ORBIT_ENV: workspace["env_name"],
                AWS_ORBIT_TEAM_SPACE: workspace["team_space"],
            },
        }
    )
    return scratch_db_name


def get_stepfunctions_waiter_config(delay: int, max_attempts: int) -> Dict[str, Any]:
    return {
        "version": 2,
        "waiters": {
            "ExecutionComplete": {
                "operation": "DescribeExecution",
                "delay": delay,
                "maxAttempts": max_attempts,
                "acceptors": [
                    {
                        "matcher": "path",
                        "expected": "SUCCEEDED",
                        "argument": "status",
                        "state": "success",
                    },
                    {
                        "matcher": "path",
                        "expected": "RUNNING",
                        "argument": "status",
                        "state": "retry",
                    },
                    {
                        "matcher": "path",
                        "expected": "FAILED",
                        "argument": "status",
                        "state": "failure",
                    },
                    {
                        "matcher": "path",
                        "expected": "TIMED_OUT",
                        "argument": "status",
                        "state": "failure",
                    },
                    {
                        "matcher": "path",
                        "expected": "ABORTED",
                        "argument": "status",
                        "state": "failure",
                    },
                ],
            },
        },
    }



import json
import logging
import time
from typing import Any, Dict, List, Optional

import boto3

from aws_orbit_sdk.common import get_workspace

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger()


def delete_crawler(crawler: str) -> None:
    """
    Deletes a glue crawler.

    Parameters
    ----------
    crawler: str
        A unique name of the Crawler

    Returns
    -------
    None
        None.

    Example
    -------
    >>> import aws_orbit_sdk.glue_catalog as glue
    >>> delete_crawler(crawler= "crawler-name")
    """
    glue = boto3.client("glue")
    glue.delete_crawler(Name=crawler)
    logger.info("existing crawler deleted")


def run_crawler(crawler: str, target_db: str, target_path: str, wait: Optional[Any] = True) -> str:
    """
    This API starts a glue crawler for the given path and will create the tables base on data found in the provided
    database. The call can wait until the crawler is done and table created.

    Parameters
    ----------
    crawler: str
        A unique name of the Crawler
    target_db: str
        The name of the target database
    target_path: str
        The S3 Path where the data for the table resides.
    wait: optional, bool
        If True, will wait until the Crawler is finished.

    Returns
    -------
    state: str
        Returns the state of the crawler after finished running and creating tables.
    Example
    -------
    >>> import aws_orbit_sdk.glue_catalog as glue
    >>> response = glue.run_crawler(crawler, target_db, target_path, wait=True)
    """
    role = get_workspace()["EksPodRoleArn"]
    glue = boto3.client("glue")
    try:
        glue.delete_crawler(Name=crawler)
        logger.info("existing crawler deleted")
    except Exception as e:
        error = str(e)
        if "EntityNotFoundException" not in error:
            logger.error(error)
        pass

    response = glue.create_crawler(
        Name=crawler,
        Role=role,
        DatabaseName=target_db,
        Targets={"S3Targets": [{"Path": target_path}]},
    )
    state = response["ResponseMetadata"]["HTTPStatusCode"]
    if state != 200:
        raise Exception("Failed to create crawler")

    glue.start_crawler(Name=crawler)

    logger.info("Crawler started...")
    state = "INIT"
    while state != "READY":
        response = glue.get_crawler(Name=crawler)
        state = response["Crawler"]["State"]
        if not wait:
            return state
        logger.info(f"Crawler in state: {state}, waiting a min... ")
        time.sleep(60)

    response = glue.get_crawler_metrics(CrawlerNameList=[crawler])
    if "CrawlerMetricsList" not in response or "TablesCreated" not in response["CrawlerMetricsList"][0]:
        raise Exception("Crawler failed to create table")

    stats = response["CrawlerMetricsList"][0]

    logger.info(stats)

    logger.info("Crawler finished creating table")
    return state


def update_teamspace_lakeformation_permissions(db_name: Optional[str] = "*") -> None:
    """
    This call will perform a scan over the provided database's tables. Base on the security selector for the given
    Team Space and base on the current column tags, the permissions will be update to allow access for permitted
    columns.

    Parameters
    -----------
    db_name: optional, str
        Name of the database for which to update permissions.

    Returns
    -------
    None
        None.

    Example
    --------
    >>> import aws_orbit_sdk.glue_catalog as glue
    >>> glue.update_teamspace_lakeformation_permissions(database_name)
    """
    workspace = get_workspace()
    lambda_client = boto3.client("lambda")

    inp = {
        "env_name": workspace["env_name"],
        "team_space": workspace["team_space"],
        "db_name": db_name,
        "role_arn": workspace["EksPodRoleArn"],
    }
    payload = json.dumps(inp)
    response = lambda_client.invoke(
        FunctionName=f"orbit-{workspace['env_name']}-authorize_lake_formation_for_role",
        InvocationType="RequestResponse",
        LogType="Tail",
        Payload=bytes(payload, "utf-8"),
    )

    if response["ResponseMetadata"]["HTTPStatusCode"] == 200:
        response_payload = json.loads(response["Payload"].read().decode("utf-8"))
        if "errorMessage" in response_payload:
            raise Exception(response_payload["errorMessage"])

    print("Lakeformation permissions have been updated")


def _update_column_parameters(table: Dict[str, Any], name: str, key: str, value: Optional[Any] = None) -> None:
    """
    Updates the column parameters of a given table.
    """
    columns = table["StorageDescriptor"]["Columns"]
    for c in columns:
        if c["Name"] == name:
            if value != None:
                if "Parameters" not in c:
                    c["Parameters"] = {}

                c["Parameters"][key] = value
            else:
                if "Parameters" in c and key in c["Parameters"]:
                    del c["Parameters"][key]


def tag_columns(
    database: str,
    table_name: str,
    key: str,
    table_tag_value: str,
    columns: Optional[List[str]] = None,
    column_tag_value: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Update the tag values for a specific table or set of columns to increase or decrease security permissions.

    Parameters
    ----------
    database: str
        Name of the database
    table_name: str
        Name of the table on which to perform the tags
    key: required, str
        Tag key
    table_tag_value: str
        Tag value to give to the entire table.
    columns: list, optional
        List of columns to update with a column tag.
    column_tag_value: str, optional
        Tag value to give to the list of columns specified in the table

    Returns
    -------
    update_table: dict
        Dictionary containing the table object metadata and new tag values.

    Examples
    --------
    >>> import aws_orbit_sdk.glue_catalog as glue
    >>> update_table = orbit_catalog_api.tag_columns(
    ...                        database='secured_database',
    ...                        table_name='table1',
    ...                        key='security-level',
    ...                        table_tag_value='sec-4')
    """
    glue = boto3.client("glue")
    response = glue.get_table(DatabaseName=database, Name=table_name)
    update_table = response["Table"]
    if "Parameters" not in update_table["StorageDescriptor"]:
        update_table["StorageDescriptor"]["Parameters"] = {}

    update_table["StorageDescriptor"]["Parameters"][key] = table_tag_value

    if "DatabaseName" in update_table:
        del update_table["DatabaseName"]
    if "CreateTime" in update_table:
        del update_table["CreateTime"]
    if "UpdateTime" in update_table:
        del update_table["UpdateTime"]
    if "CreatedBy" in update_table:
        del update_table["CreatedBy"]
    if "IsRegisteredWithLakeFormation" in update_table:
        del update_table["IsRegisteredWithLakeFormation"]
    if columns != None:
        for c in columns:
            _update_column_parameters(update_table, c, key, column_tag_value)
    glue.update_table(DatabaseName=database, TableInput=update_table)
    return update_table


def untag_columns(
    database: str,
    table_name: Optional[str] = None,
    columns: Optional[List[str]] = None,
    key: Optional[str] = None,
    table_tag: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Removes tags from a specified column or table, or all the tables in a given database.

    Note
    ----
    If a tag value is specified, the tag key gets updated to the ne tag value instead of the key being deleted.

    Parameter
    ---------
    database: str
        Name of the database
    table_name: str
        Name of the table on which to untag if specified
    columns: list
        List of columns to untag if specified
    key: str
        Tag key to remove
    table_tag: str
        Tag value to give to the entire table.

    Return
    ------
    update_table: dict
        Dictionary containing the table object metadata after being untagged.

    Example
    -------
    >>> import aws_orbit_sdk.glue_catalog as glue
    >>> orbit_catalog_api.untag_columns(database='secured_database',key='security-level')
    """
    glue = boto3.client("glue")
    table_names = []
    if table_name != None:
        table_names = [table_name]
    else:
        tables = glue.get_tables(DatabaseName=database)["TableList"]

        for table in tables:
            table_names.append(table["Name"])

    update_table = {}
    for table_name in table_names:
        response = glue.get_table(DatabaseName=database, Name=table_name)
        update_table = response["Table"]
        if "Parameters" not in update_table["StorageDescriptor"]:
            update_table["StorageDescriptor"]["Parameters"] = {}

        if table_tag != None:
            update_table["StorageDescriptor"]["Parameters"][key] = table_tag
        else:
            if key in update_table["StorageDescriptor"]["Parameters"]:
                del update_table["StorageDescriptor"]["Parameters"][key]

        if "DatabaseName" in update_table:
            del update_table["DatabaseName"]
        if "CreateTime" in update_table:
            del update_table["CreateTime"]
        if "UpdateTime" in update_table:
            del update_table["UpdateTime"]
        if "CreatedBy" in update_table:
            del update_table["CreatedBy"]
        if "IsRegisteredWithLakeFormation" in update_table:
            del update_table["IsRegisteredWithLakeFormation"]

        if columns == None:
            cols = table["StorageDescriptor"]["Columns"]
            columns = []
            for c in cols:
                columns.append(c["Name"])

        for c in columns:
            _update_column_parameters(update_table, c, key)

        logger.info(f"untagging table {table_name}")
        glue.update_table(DatabaseName=database, TableInput=update_table)
    return update_table


def getCatalogAsDict(database: Optional[str] = None) -> Dict:
    """
    Get Data Catalog of a specific Database

    Parameters
    ----------
    database : str
        Name of database to catalog.

    Returns
    -------
     schema : IPython.core.display.JSON
        An IPython JSON display representing the schema metadata for a database

    Example
    --------
    >>> from aws.utils.notebooks.database import AthenaUtils
    >>> from aws.utils.notebooks.json import display_json
    >>> AthenaUtils.getCatalog(database="my_database")
    """
    glue = boto3.client("glue")
    schemas: List[Dict[str, Any]] = []
    response = glue.get_databases()
    key: int = 0
    for db in response["DatabaseList"]:
        key += 1
        database = {"title": db["Name"], "key": str(key), "qname": db["Name"], "children": [], "_class": "database"}
        schemas.append(database)
        response = glue.get_tables(DatabaseName=db["Name"], MaxResults=50)
        for t in response["TableList"]:
            key += 1
            table = dict()
            table["title"] = t["Name"]
            table["key"] = str(key)
            table["location"] = t["StorageDescriptor"]["Location"]
            table["children"] = []
            table["key"] = str(key)
            table["_class"] = "table"
            table["db"] = db["Name"]
            table["table"] = t["Name"]
            database["children"].append(table)
            for c in t["StorageDescriptor"]["Columns"]:
                col = dict()
                key += 1
                table["children"].append(col)
                col["title"] = c["Name"]
                col["type"] = c["Type"]
                col["key"] = str(key)
                col["_class"] = "column"
                col["db"] = db["Name"]
                col["table"] = t["Name"]
    return schemas



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
from typing import Any, Dict, List, Optional, Union, cast

import boto3

_logger = logging.getLogger()

PROFILE_TYPE = Dict[str, Union[str, Dict[str, Any]]]
PROFILES_TYPE = List[PROFILE_TYPE]


class TeamConstants:
    def __init__(self, username=None) -> None:
        self.env_name = os.environ["AWS_ORBIT_ENV"]
        self.team_name = os.environ["AWS_ORBIT_TEAM_SPACE"]
        self.region = os.environ["AWS_DEFAULT_REGION"]
        self.account_id = str(boto3.client("sts").get_caller_identity().get("Account"))
        if username:
            self.username = username
        else:
            self.username = (
                os.environ["JUPYTERHUB_USER"]
                if "JUPYTERHUB_USER" in os.environ
                else os.environ["USERNAME"]
                if "USERNAME" in os.environ
                else None
            )
        self.pvc_name_template = f"orbit-{self.username}"

    def image_pull_policy(self) -> str:
        return "Always"

    def gid(self) -> int:
        return 100

    def uid(self, grant_sudo: bool) -> int:
        if grant_sudo:
            return 0
        return 1000

    def node_selector(self, node_type: str) -> List[Dict[str, str]]:
        node_selector = {"orbit/usage": "teams"}
        if node_type == "ec2":
            node_selector["orbit/node-type"] = "ec2"
        return node_selector

    def volume_mounts(self) -> List[Dict[str, Any]]:
        mounts = [{"mountPath": "/efs", "name": "efs-volume"}]
        return mounts

    def volumes(self) -> List[Dict[str, Any]]:
        vols = [
            {"name": "efs-volume", "persistentVolumeClaim": {"claimName": "jupyterhub"}},
        ]
        return vols

    def default_image(self) -> str:
        return f"{self.account_id}.dkr.ecr.{self.region}.amazonaws.com/orbit-{self.env_name}-{self.team_name}:latest"

    def default_spark_image(self) -> str:
        return (
            f"{self.account_id}.dkr.ecr.{self.region}.amazonaws.com/orbit-{self.env_name}-{self.team_name}-spark:latest"
        )

    def env(self) -> Dict[str, str]:
        env = dict()
        env["AWS_ORBIT_ENV"] = self.env_name
        env["AWS_ORBIT_TEAM_SPACE"] = self.team_name
        env["AWS_STS_REGIONAL_ENDPOINTS"] = "regional"
        return env

    def default_profiles(self) -> PROFILES_TYPE:
        return [
            {
                "display_name": "Nano",
                "slug": "nano",
                "description": "1 CPU + 1G MEM",
                "kubespawner_override": {"cpu_guarantee": 1, "cpu_limit": 1, "mem_guarantee": "1G", "mem_limit": "1G"},
            },
            {
                "display_name": "Micro",
                "slug": "micro",
                "description": "2 CPU + 2G MEM",
                "kubespawner_override": {
                    "cpu_guarantee": 2,
                    "cpu_limit": 2,
                    "mem_guarantee": "2G",
                    "mem_limit": "2G",
                },
                "default": True,
            },
            {
                "display_name": "Small",
                "slug": "small",
                "description": "4 CPU + 8G MEM",
                "kubespawner_override": {
                    "cpu_guarantee": 4,
                    "cpu_limit": 4,
                    "mem_guarantee": "8G",
                    "mem_limit": "8G",
                },
            },
        ]

    def deployed_profiles(self) -> PROFILES_TYPE:
        ssm = boto3.Session().client("ssm")
        ssm_parameter_name: str = f"/orbit/{self.env_name}/teams/{self.team_name}/context"
        json_str: str = ssm.get_parameter(Name=ssm_parameter_name)["Parameter"]["Value"]

        team_manifest_dic = cast(PROFILES_TYPE, json.loads(json_str))
        deployed_profiles: PROFILES_TYPE = []
        if team_manifest_dic.get("profiles"):
            deployed_profiles = team_manifest_dic["profiles"]
        if len(deployed_profiles) == 0:
            deployed_profiles = self.default_profiles()

        return deployed_profiles

    def team_profiles(self) -> PROFILES_TYPE:
        ssm = boto3.Session().client("ssm")

        default_profiles = self.deployed_profiles()

        ssm_parameter_name: str = f"/orbit/{self.env_name}/teams/{self.team_name}/user/profiles"
        json_str: str = ssm.get_parameter(Name=ssm_parameter_name)["Parameter"]["Value"]

        user_profiles: PROFILES_TYPE = cast(PROFILES_TYPE, json.loads(json_str))
        default_profiles.extend(user_profiles)
        _logger.debug("profiles:%s", default_profiles)
        return default_profiles

    def default_profile(self) -> PROFILE_TYPE:
        for p in self.team_profiles():
            if "default" in p and (p["default"] == "True" or p["default"] == True):
                return p
        return None

    def profile(self, name: str) -> PROFILE_TYPE:
        for p in self.team_profiles():
            if p["slug"] == name:
                return p
        return None

    def init_containers(self, image):
        return [
            {
                "name": "take-ebs-dir-ownership",
                "image": image,
                "command": ["sh", "-c", "sudo chmod -R 777 /ebs"],
                "securityContext": {"runAsUser": 0},
                "volumeMounts": [{"mountPath": "/ebs", "name": "ebs-volume"}],
            }
        ]

    def life_cycle_hooks(self):
        return {}
        # return {"postStart": {"exec": {"command": ["/bin/sh", "/efs/shared/bootstrap.sh"]}}}

        # return {"postStart": {"exec": {"command": ["/bin/sh", "/etc/jupyterhub/bootstrap.sh", "2>&1","|","tee /efs/shared/bootstrap.log"]}}}

    def annotations(self) -> Dict[str, str]:
        anno = {"AWS_ORBIT_TEAM_SPACE": self.team_name, "AWS_ORBIT_ENV": self.env_name}

    def storage_class(self) -> str:
        return f"ebs-{self.team_name}-gp2"

    def ebs_access_mode(self) -> List[str]:
        return ["ReadWriteOnce"]



import glob
import json
import logging
import os
import subprocess
from os.path import expanduser
from pathlib import Path
from tempfile import mkstemp
from typing import Any, Dict, List, Optional, Union

import boto3
import IPython.display
from IPython.display import JSON

from aws_orbit_sdk.common import get_properties, get_workspace, split_s3_path

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger()


def display_json(doc: Dict[str, Any], root: Optional[str] = "root") -> IPython.core.display.JSON:
    """
    Create a JSON display object given raw JSON data.

    Parameters
    ----------
    doc : JSON dict or list
        Raw JSON data to display.
    root : str, optional
        The name of the root element of the JSON tree (default 'root').

    Returns
    -------
    IPython display of JSON data : IPython.core.display.JSON
        An IPython JSON display representing the data.

    Example
    -------
    >>> from aws.utils.notebooks import json  as json_utils
    >>> json_utils.display_json(doc=my_schemas ,root="database")
    """

    return JSON(doc)


def write_json(doc: Any, path: str) -> None:
    """
    Write data to a s3 bucket path.

    Parameters
    ----------
    doc : Any (Object Data)
        Data to write to a s3 bucket.

    path : str
        s3 Bucket path for where to write data.

    Returns
    -------
    None
        None.

    Example
    -------
    >>> from aws.utils.notebooks import json  as json_utils
    >>> data = {
    ...     'id' : [1]
    ...     'name': ['Noah']
    ...     }
    >>> write_json(doc=data, path='testbucket123')
    """

    s3 = boto3.client("s3")
    if path.startswith("s3://"):
        (bucket, key) = split_s3_path(path)
        s3.put_object(Body=doc, Bucket=bucket, Key=key)

    dir = os.path.dirname(path)
    Path(dir).mkdir(parents=True, exist_ok=True)

    with open(path, "w") as outfile:
        json.dump(doc, outfile)


def run_schema_induction(
    data_path: str,
    table_name: str,
    s3_location: str,
    root_definition_name: str,
    is_array: Optional[bool] = True,
) -> Dict[str, Dict[str, str]]:
    """
    Calls on helper functions to run Schema Induction with given user arguments and returns ddl and schema metadata.

    Parameters
    ----------
    data_path : str
        An input json file path.
    table_name : str
        Table name to use when creating DDL.
    s3_location : str
        Table location to use when creating DDL.
    root_definition_name : str
        The root directory name of the data.
    is_array : bool, optional
        Is the document a json array.

    Returns
    -------
    ddl : str
        SQL ddl statement to create new external table with given metadata
    schema : dict
        Schema metadata stored as a json for the specified table.

    Example
    -------
    >>> from aws.utils.notebooks import json  as json_utils
    >>>table = json_utils.run_schema_induction(data_path=f's3://{target_bucket}/Claim-1/data.json',
    ...              table_name='users.table',
    ...              s3_location=f's3://{target_bucket}/Claim-1/',
    ...              root_definition_name='Claim',
    ...              is_array=True)
    >>> json_utils.display_json(table['schema'])
    >>> ddl = table['ddl']
    """

    logger.info("Start induction process for " + table_name)

    args = [
        "-i",
        data_path,
        "-c",
        "ec2",
        "-t",
        table_name,
        "--location",
        s3_location,
        "--root",
        root_definition_name,
    ]
    if is_array:
        args.append("-a")

    ret = run_schema_induction_args(args)

    logger.info("Finish induction process for " + table_name)

    return ret


def run_process(args: Union[str, List[str]]) -> None:
    """
    Helper function called by run_schema_induction_args() to log outputs and errors.

    Parameters
    ----------
    args : Union[str, List[str]]
        User arguments used for the schema induction process to run.

    Returns
    -------
    None
        None.
    """
    try:
        completed = subprocess.run(
            args,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as err:
        logger.error(f"ERROR: {err}")
    else:
        out = completed.stdout.decode("utf-8")
        err = completed.stderr.decode("utf-8")
        if len(out) > 0:
            logger.info(out)
        if len(err):
            logger.error(err)


def run_schema_induction_args(user_args: Union[str, List[str]]) -> Dict[str, Dict[str, str]]:
    """
    Calls on run_process to run Schema Induction with given user arguments gets ddl and schema metadata for a specified
    table.

    Parameters
    ----------
    user_args : Union[str, List[str]]
        User parameters including ( data_path, instance type, table_name, s3_location, and root definition name)

    Returns
    -------
    ddl : str
        SQL ddl statement to create new external table with given metadata.
    schema :
        Schema metadata stored as a json for the specified table.


    Examples
    --------
    >>>from aws.utils.notebooks import json  as json_utils
    >>>args = [ "-i data_path -c ec2 -t table_name --location s3_location --root ClaimData"]
    >>>args.append("-a")
    >>>table = json_utils.run_schema_induction_args(args)
    """

    ddlFd, ddl_path = mkstemp()
    schemaFd, schema_path = mkstemp()

    home = expanduser("~")

    jar = glob.glob(f"{home}/orbit/java/schema-induction*.jar")[0]

    args = [
        "/opt/jdk-13.0.1/bin/java",
        "-jar",
        f"{jar}",
        "-s",
        schema_path,
        "-d",
        ddl_path,
    ]
    args.extend(user_args)

    run_process(args)

    ddl = readFile(ddl_path, ddlFd)
    schema = readFile(schema_path, schemaFd)

    return {"ddl": ddl, "schema": schema}


def readFile(path: str, fd: int) -> Any:
    """
    Read File Contents located at a specified path.

    Parameters
    ----------
    path : str
        File path to locate file to read.
    fd : int
        Associated file descriptor

    Returns
    -------
    content : Any
        File contents

    Example
    -------
    >>>from aws.utils.notebooks import json  as json_utils
    >>>path = "/home/ihritik/Desktop/file2.txt"
    >>>fd = os.open( "foo.txt", os.O_RDONLY)
    >>>data = json_utils.readFile(path=path,fd=fd)
    """

    file = open(path, "r")
    content = file.read()
    file.close()
    os.close(fd)
    os.remove(path)
    return content



import argparse
import sys
from typing import Any, Dict, Optional

import IPython.display
import sql.connection
from IPython import get_ipython
from IPython.core.magic import Magics, cell_magic, line_magic, magics_class, needs_local_scope
from IPython.display import JSON


def exception_handler(exception_type, exception, traceback):
    print("%s: %s" % (exception_type.__name__, exception), file=sys.stderr)


class ArgumentParserNoSysExit(argparse.ArgumentParser):
    def error(self, message):
        self.print_help(sys.stderr)
        raise Exception(message)


@magics_class
class DatabaseMagics(Magics):
    """
    Database functions, inherited by both Redshift and Athena classes.
    """

    def __init__(self, shell, database_utils):
        super().__init__(shell)
        self.ip = get_ipython()
        self.database_utils = database_utils


@magics_class
class RedshiftMagics(DatabaseMagics):
    """
    Redshift Custom Magic functions used to easily connect and work with databases.

    Methods
    -------
    %connect_to_redshift:
        Start Redshift connection
    %connect_to_external_redshift:
        Start Redshift connection
    %delete_redshift_cluster:
        Deletes a redshift cluster.
    %catalog:
        Get Glue Catalog metadata of a specific Database table.
    %create_external_schema:
        Creates external schema if it does not already exist from a glue data catalog database.
    %%ddl:
        Executes a SQL ddl statement.
    %%create_external_table:
        Create external table in S3 and Glue Catalog.
    """

    def __init__(self, shell, database_utils):
        super().__init__(shell, database_utils)

    @line_magic
    def connect_to_redshift(self, line: str) -> None:
        """
        Start Redshift connection

        Parameters
        ----------
        '-cluster' : str
            Specify cluster name
        '-start' : bool, optional
            Start cluster if it does not exists (default False)
        '-reuse' : bool, optional
            Reuse cluster if cluster name exists (default True)
        '-func' : str, optional
            Other redshift arguments(e.g. max_concurrency_scaling_clusters, etc.)(default None).


        Returns
        -------
        None
            None.

        Examples
        --------
        >>> %connect_to_redshift -cluster db-test -reuse -start -func Standard Nodes=3
        """

        if len(line) == 0:
            print("must provide redshift cluster name")
            return

        parser = ArgumentParserNoSysExit(description="start Redshift connection")
        parser.add_argument("-cluster", required=True, help="specify cluster name")
        parser.add_argument("-start", action="store_true", help="start cluster if not exists")
        parser.add_argument("-reuse", action="store_true", help="reuse cluster if exists")
        parser.add_argument("-func", nargs="*", default=None, help="specify cluster function name")
        try:
            args = parser.parse_args(line.strip().split(" "))
            clusterArgs = dict()
            if args.func != None:
                clusterArgs["redshift_start_function"] = args.func[0]
                for arg in args.func[1:]:
                    keypair = arg.split("=")
                    clusterArgs[keypair[0]] = keypair[1]

            connProp = self.database_utils.connect_to_redshift(
                cluster_name=args.cluster,
                reuseCluster=args.reuse,
                startCluster=args.start,
                clusterArgs=clusterArgs,
            )
            dbUrl = connProp["db_url"]
            self.ip.run_line_magic("sql", dbUrl)
            print("connected!")
        except Exception as e:
            print("Error!")
            print(str(e))

    @line_magic
    def connect_to_external_redshift(self, line: str) -> None:
        """
        Start Redshift connection

        Parameters
        ----------
        '-cluster' : str
            specify cluster name
        '-DbName' : str
            specify the Redshift DB name to connect to.
        '-DbUser' : str
            specify DB the Redshift User name to connect with.
        '-lambdaName' : str, optional
            For new cluster/mandatory for existing cluster connection, the lambda name which is responsible to get the
                cluster credentials.

        Returns
        -------
        None
            None.

        Example
        -------
        >>> %connect_to_external_redshift -cluster db-test -DbName mydatabase -DbUser my_user
        """

        if len(line) == 0:
            print("must provide redshift cluster name")
            return

        parser = ArgumentParserNoSysExit(description="start Redshift connection")
        parser.add_argument("-cluster", required=True, help="specify cluster name")
        parser.add_argument("-DbName", required=True, help="specify DB name")
        parser.add_argument("-DbUser", required=True, help="specify DB user")
        parser.add_argument("-lambdaName", default=None, help="specify credentials lambda name")
        try:
            args = parser.parse_args(line.strip().split(" "))
            connProp = self.database_utils.get_connection_to_redshift(
                clusterIdentifier=args.cluster,
                DbName=args.DbName,
                DbUser=args.DbUser,
                lambdaName=args.lambdaName,
            )
            dbUrl = connProp["db_url"]
            self.ip.run_line_magic("sql", dbUrl)
            print("connected!")
        except Exception as e:
            print("Error!")
            print(str(e))

    @line_magic
    def delete_redshift_cluster(self, line: str) -> None:
        """
        Deletes a redshift cluster.

        Parameters
        ----------
        '-cluster' : str
            The Redshift cluster name

        Returns
        -------
        None
            None.

        Example
        --------
        >>> %delete_redshift_cluster -cluster db-test
        """

        if len(line) == 0:
            print("must provide redshift cluster name")
            return

        parser = ArgumentParserNoSysExit(description="start Redshift connection")
        parser.add_argument("-cluster", required=True, help="specify cluster name")
        try:
            args = parser.parse_args(line.strip().split(" "))
            self.database_utils.delete_redshift_cluster(args.cluster)
        except Exception as e:
            print("Error!")
            print(str(e))

    @needs_local_scope
    @cell_magic
    def ddl(self, line: str, cell: str, local_ns: Optional[Dict[str, str]] = None) -> None:
        """
        Executes a SQL ddl statement.

        Parameters
        ----------
        cell : str
            The data descrption language statement to execute in SQL.
        local_ns : str
            The namespace to load into IPython user namespace

        Returns
        -------
        None
            None.

        Examples
        --------
        >>>%%ddl
        >>>drop table if exists mydatabase.Table3
        """

        self.database_utils.execute_ddl(cell, local_ns)

    @line_magic
    def catalog(self, line: str) -> IPython.core.display.JSON:
        """
        Get Glue Catalog metadata of a specific Database table.

        Parameters
        ----------
        -s : str, optional
            Name of schema to retrieve Glue Catalog for (default looks at all schema with given table names in the
            current database)

        -t : str, optional
            Name of table to retrieve Glue Catalog for (default looks at all tables with given schema names in the
            current database)

        Returns
        -------
        schema : IPython.core.display.JSON
            An IPython JSON display representing the schema metadata for the table(s) / database

        Example
        --------
        >>> %catalog -s database_ext_schema
        """

        if len(line) == 0:
            return self.database_utils.getCatalog()
        parser = ArgumentParserNoSysExit(description="display all databases, tables and columns")
        parser.add_argument("-s", nargs="?", default=None, help="specify external schema name")
        parser.add_argument("-t", nargs="?", default=None, help="specify external table name")

        try:
            args = parser.parse_args(line.strip().split(" "))
            return self.database_utils.getCatalog(schema_name=args.s, table_name=args.t)
        except Exception as e:
            print("Error!")
            print(str(e))

    @line_magic
    def create_external_schema(self, line: str) -> None:
        """
        Creates external schema if it does not already exist from a glue data catalog database.

        Parameters
        ----------
        -s : str
            Name of the external schema that will be created.

        -g : str
            Name of the glue database for creating the schema.

        Returns
        -------
        None
            None.

        Example
        --------
        >>> %create_external_schema -s users -g users
        """
        if len(line) == 0:
            return self.database_utils.getCatalog()
        parser = ArgumentParserNoSysExit(description="create external schema in Redshift to point to Glue Catalog")
        parser.add_argument("-s", nargs="?", default=None, help="specify external schema name")
        parser.add_argument("-g", nargs="?", default=None, help="specify glue database")

        try:
            args = parser.parse_args(line.strip().split(" "))

            self.database_utils.create_external_schema(schema_name=args.s, glue_database=args.g)
        except Exception as e:
            print("Error!")
            print(str(e))

    @cell_magic
    def create_external_table(
        self, line: str, cell: str, local_ns: Optional[Dict[str, str]] = None
    ) -> IPython.core.display.JSON:
        """
        Create external table in S3 and Glue Catalog

        Parameters
        ----------
        -g : str
            Name of database with existing data and schema.
        -t :
            The name of the table to be created.
        -f : str, optional
            The file format for data files (default Parquet format).
        -l : str, optional
            The path to the Amazon S3 bucket or folder that contains the data files or a manifest file that contains a
            list of Amazon S3 object paths (used only if no database has no location).
        -u : str, optional
            Specify additional unload properties for when creating new table.

        Returns
        -------
        s : IPython.core.display.JSON
            An IPython JSON display representing the schema metadata for the table(s) / database

        Example
        --------
        >>> %%create_external_table -g users -t myQuery1
        >>> select c.id as id1,c.status from users.Table1 as c
        """
        if len(line) == 0:
            return self.database_utils.getCatalog()
        parser = ArgumentParserNoSysExit(description="create external table in S3 and Glue Catalog")
        parser.add_argument("-g", help="specify glue database name")
        parser.add_argument("-t", help="specify table name")
        parser.add_argument(
            "-f",
            nargs="?",
            default="parquet",
            choices=["parquet", "csv"],
            help="specify file format",
        )
        parser.add_argument(
            "-l",
            nargs="?",
            default=None,
            help="specify s3 location or leave out to use Glue DB location",
        )
        parser.add_argument("-u", nargs="?", default="", help="other unload properties")

        try:
            args = parser.parse_args(line.strip().split(" "))

            return self.database_utils.create_external_table(
                select=cell,
                database_name=args.g,
                table_name=args.t,
                format=args.f,
                s3_location=args.l,
                options=args.u,
            )
        except Exception as e:
            print("Error!")
            print(str(e))

    #
    # def apt_completers(self, event):
    #     """ This should return a list of strings with possible completions.
    #
    #     Note that all the included strings that don't start with event.symbol
    #     are removed, in order to not confuse readline.
    #     """
    #     return ['update', 'upgrade', 'install', 'remove']
    #
    #     completerlib.quick_completer('SELECT', ['FROM','WHERE', 'GROUPBY'])
    #
    # magics_class


@magics_class
class AthenaMagics(DatabaseMagics):
    """
    Collection of Athena Magics functions used to easily connect and work with databases.

    Methods
    -------
    %connect_to_athena:
        Connect Athena to an existing database
    %catalog:
        Get Data Catalog and display all databases, tables and columns

    """

    def __init__(self, shell, database_utils):
        super().__init__(shell, database_utils)

    @line_magic
    def connect_to_athena(self, line: str) -> None:
        """
        Connect Athena to an existing database

        Parameters
        ----------
        -database : str
            Name of the glue database name.

        Returns
        -------
        None
            None.

        Example
        --------
        >>> %connect_to_athena -database users_data
        """

        if len(line) == 0:
            print("must provide an athena database name")
            return

        parser = ArgumentParserNoSysExit(description="start Redshift connection")
        parser.add_argument("-database", required=True, help="specify database name")
        try:
            args = parser.parse_args(line.strip().split(" "))
            connProp = self.database_utils.get_connection_to_athena(args.database)
            dbUrl = connProp["db_url"]
            self.ip.run_line_magic("sql", dbUrl)
            print("connected!")
        except Exception as e:
            print("Error!")
            print(str(e))

    @line_magic
    def close_athena(self, line: str) -> None:
        """
        Close all Athena connections opened in the notebook

        Returns
        -------
        None
            None.

        Example
        --------
        >>> %close_athena
        """
        sqlmagic_connections = sql.connection.Connection.connection
        for conn_key in list(sqlmagic_connections.keys()):
            conn = sqlmagic_connections[conn_key]
            conn.session.close()
        print("Connections closed")

    @line_magic
    def catalog(self, line: str) -> IPython.core.display.JSON:
        """
        Get Data Catalog and display all databases, tables and columns

        Parameters
        ----------
        -database : str, optional
            Name of database to catalog (default looks at all databases).

        Returns
        -------
        schema : IPython.core.display.JSON
            An IPython JSON display representing the schema metadata for a database

        Example
        --------
        >>> %catalog -database users_data
        """
        if len(line) == 0:
            return self.database_utils.getCatalog()
        parser = ArgumentParserNoSysExit(description="display all databases, tables and columns")
        parser.add_argument(
            "-database",
            nargs="?",
            default=None,
            required=True,
            help="specify database name ",
        )

        try:
            args = parser.parse_args(line.strip().split(" "))
            return self.database_utils.getCatalog(database=args.database)
        except Exception as e:
            print("Error!")
            print(str(e))



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import sys
import time
import urllib
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

import boto3
import botocore
import pandas as pd
from kubernetes import config as k8_config
from kubernetes import watch as k8_watch
from kubernetes.client import *
from kubespawner.objects import make_pod, make_pvc
from slugify import slugify

from aws_orbit_sdk.common import get_properties, get_stepfunctions_waiter_config
from aws_orbit_sdk.common_pod_specification import TeamConstants

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)
_logger = logging.getLogger()

MANIFEST_PLUGIN_TYPE = Dict[str, Union[str, Dict[str, Any]]]
MANIFEST_PROPERTY_MAP_TYPE = Dict[str, Union[str, Dict[str, Any]]]
MANIFEST_FILE_TEAM_TYPE = Dict[str, Union[str, int, None, List[MANIFEST_PROPERTY_MAP_TYPE], List[str]]]
MANIFEST_TEAM_TYPE = Dict[str, Union[str, int, None, List[MANIFEST_PLUGIN_TYPE]]]
MANIFEST_PROPERTY_MAP_TYPE = Dict[str, Union[str, Dict[str, Any]]]

__CURRENT_TEAM_MANIFEST__: MANIFEST_TEAM_TYPE = None
__CURRENT_ENV_MANIFEST__: MANIFEST_TEAM_TYPE = None


def read_team_manifest_ssm(env_name: str, team_name: str) -> Optional[MANIFEST_TEAM_TYPE]:
    parameter_name: str = f"/orbit/{env_name}/teams/{team_name}/manifest"
    _logger.debug("Trying to read manifest from SSM parameter (%s).", parameter_name)
    client = boto3.client("ssm")
    try:
        json_str: str = client.get_parameter(Name=parameter_name)["Parameter"]["Value"]
    except client.exceptions.ParameterNotFound:
        _logger.debug("Team %s Manifest SSM parameter not found: %s", team_name, parameter_name)
        return None
    _logger.debug("Team %s Manifest SSM parameter found.", team_name)
    return cast(MANIFEST_TEAM_TYPE, json.loads(json_str))


def get_parameter(client, name: str) -> Dict[str, Any]:
    try:
        json_str: str = client.get_parameter(Name=name)["Parameter"]["Value"]
    except botocore.errorfactory.ParameterNotFound as e:
        _logger.error("failed to read parameter %s", name)
        raise
    return cast(Dict[str, Any], json.loads(json_str))


def load_env_context_from_ssm(env_name: str) -> Optional[MANIFEST_TEAM_TYPE]:
    ssm = boto3.client("ssm")
    context_parameter_name: str = f"/orbit/{env_name}/context"
    context = get_parameter(ssm, name=context_parameter_name)
    return cast(MANIFEST_TEAM_TYPE, context)


def load_team_context_from_ssm(env_name: str, team_name: str) -> Optional[MANIFEST_TEAM_TYPE]:
    ssm = boto3.client("ssm")
    context_parameter_name: str = f"/orbit/{env_name}/teams/{team_name}/context"
    context = get_parameter(ssm, name=context_parameter_name)
    return cast(MANIFEST_TEAM_TYPE, context)


def get_execution_history(notebookDir: str, notebookName: str) -> pd.DataFrame:
    """
     Get Notebook Execution History

     Parameters
     ----------
     notebookDir: str
         Name of notebook directory.
     notebookName: str
         Name of notebook.

     Returns
     -------
     df: pd.DataFrame
         Notebook execution history.

     Example
     --------
    >>> from aws_orbit_sdk import controller
    >>> controller.get_execution_history(notebookDir="notebook-directory", notebookName='mynotebook')
    """
    props = get_properties()
    return _get_execution_history_from_local(notebookDir, notebookName, props)


def _get_execution_history_from_local(notebook_basedir: str, src_notebook: str, props: dict) -> pd.DataFrame:
    """
    Get Notebook Execution History from EFS
    """

    home = str(Path.home())
    nb_name = Path(src_notebook).stem
    notebook_dir = os.path.join(home, notebook_basedir, nb_name)

    executions = []
    for nb in Path(notebook_dir).glob("*.ipynb"):
        if not nb.is_file():
            continue
        executions.append((str(nb), datetime.fromtimestamp(nb.stat().st_mtime), notebook_dir))

    if not executions:
        _logger.info(f"No output notebooks founds at: {notebook_dir}")

    df = pd.DataFrame(executions, columns=["relativePath", "timestamp", "path"])
    return df


def _get_execution_history_from_s3(notebookBaseDir: str, srcNotebook: str, props: str) -> pd.DataFrame:
    """
    Get Notebook Execution History from s3
    """
    s3 = boto3.client("s3")

    notebookDir = os.path.join(notebookBaseDir, srcNotebook.split(".")[0])

    path = "{}/output/notebooks/{}/".format(props["AWS_ORBIT_TEAM_SPACE"], notebookDir)
    executions = []
    objects = s3.list_objects_v2(Bucket=props["AWS_ORBIT_S3_BUCKET"], Prefix=path)
    if "Contents" in objects.keys():
        for key in s3.list_objects_v2(Bucket=props["AWS_ORBIT_S3_BUCKET"], Prefix=path)["Contents"]:
            if key["Key"][-1] == "/":
                continue
            notebookName = os.path.basename(key["Key"])
            path = key["Key"]
            arg = urllib.parse.quote(path)
            s3path = "s3://{}/{}".format(props["AWS_ORBIT_S3_BUCKET"], path)
            #     link = '<a href="{}" target="_blank">{}</a>'.format(site + arg,"open")
            timestamp = key["LastModified"]
            executions.append((notebookName, timestamp, s3path))
    else:
        _logger.info("No output notebooks founds at: s3://{}/{}".format(props["AWS_ORBIT_S3_BUCKET"], path))

    df = pd.DataFrame(executions, columns=["relativePath", "timestamp", "s3path"])
    return df


def _get_invoke_function_name() -> Any:
    """
    Get invoke function Name.

    Returns
    -------
    Function Name.
    """
    props = get_properties()
    functionName = f"orbit-{props['AWS_ORBIT_ENV']}-{props['AWS_ORBIT_TEAM_SPACE']}-container-runner"
    return functionName


def run_python(taskConfiguration: dict) -> Any:
    """
    Runs Python Task

    Parameters
    ----------
    taskConfiguration : dict
        A task definition to execute.

        tasks : lst
            A list of python task definition to run.
        module : str
           The python module to run (without .py ext).
        functionName : str
                The python function to start the execution.
        sourcePaths : lst
              A list of s3 python source paths used for importing packages or modules into the application.
        params : dict
             A list of parameters for this task to override the notebook parameters.
        compute : optional, dict
              A list of runtime parameters to control execution.
        container : dict
               A list of parameters to control container execution.
        p_concurrent : str
              The number of parallel threads inside the container that will execute notebooks.
        env_vars : optional, list
              A list of environment parameters to pass to the container.

    Returns
    -------
    response: Any
        Response for the function or an error object.

    Example
    --------
    >>> import aws_orbit_sdk.controller as controller
    >>> response = controller.run_python(
    ...     taskConfiguration = {
    ...         "tasks":  [
    ...             {
    ...                 "module": "pyspark.run_pyspark_local",
    ...                 "functionName": "run_spark_job",
    ...                 "sourcePaths": ["DataScienceRepo/samples/python"],
    ...                  "params": {
    ...                            "bucket": "users-env2",
    ...                          "p2": 'bar'
    ...                  }
    ...              }
    ...          ],
    ...          "compute": {
    ...             "container" : {
    ...                 "p_concurrent": "4"
    ...          },
    ...           "env_vars": [
    ...                      {
    ...                         'name': 'cluster_name',
    ...                          'value': clusterName
    ...                      }
    ...          ]
    ...      })
    """
    taskConfiguration["task_type"] = "python"
    if "compute_type" not in taskConfiguration or taskConfiguration["compute_type"] == "eks":
        return _run_task_eks(taskConfiguration)
    else:
        raise RuntimeError("Unsupported compute_type '%s'", taskConfiguration["compute_type"])


def run_notebooks(taskConfiguration: dict) -> Any:
    """
    Runs Notebooks Tasks

    Parameters
    ----------
    taskConfiguration : dict
        A task definition to execute.
        notebooks : lst
            A list of notebook task definition to run.
        notebookName : str
            The filename of the notebook.
        notebookName : str
            The relative path to the notebook file starting at the repository root.
        targetPath : str
             The target S3 directory where the output notebook and all related output will be generated.
        params : dict
             A list of parameters for this task to override the notebook parameters.
        compute : optional, dict
              A list of runtime parameters to control execution.
        container : dict
               A list of parameters to control container execution.
        p_concurrent : str
              The number of parallel processes inside the container that will execute notebooks.
        sns.topic.name : str
              A name of a topic to which messages are sent on task completion or failure.
        env_vars : optional, lst
             A list of environment parameters to pass to the container.

    Returns
    -------
    response: Any
        Response for the function or an error object.

    Example
    --------
    >>> import aws_orbit_sdk.controller as controller
    >>> response = controller.run_notebooks(
    ...     taskConfiguration = {
    ...         "notebooks":  [ {
    ...           "notebookName": "Example-2-Extract-Files.ipynb",
    ...           "sourcePath": "samples/notebooks/A-LakeCreator",
    ...           "targetPath": "tests/createLake",
    ...           "params": {
    ...             "bucketName": bucketName,
    ...             "zipFileName": file,
    ...             "targetFolder": extractedFolder
    ...           },
    ...           ...
    ...         },
    ...         "compute": {
    ...             "container" : {
    ...                 "p_concurrent": "4",
    ...             },
    ...             "env_vars": [
    ...                         {
    ...                             'name': 'cluster_name',
    ...                             'value': clusterName
    ...                         }
    ...             ],
    ...             "sns.topic.name": 'TestTopic',
    ...         }
    ... )
    """
    taskConfiguration["task_type"] = "jupyter"
    if "compute_type" not in taskConfiguration or taskConfiguration["compute_type"] == "eks":
        return _run_task_eks(taskConfiguration)
    else:
        raise RuntimeError("Unsupported compute_type '%s'", taskConfiguration["compute_type"])


def list_team_running_jobs():
    return list_running_jobs(True)


def list_my_running_jobs():
    return list_running_jobs(False)


def list_running_jobs(team_only: bool = False):
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    load_kube_config()
    username = os.environ.get("JUPYTERHUB_USER", os.environ.get("USERNAME"))
    api_instance = BatchV1Api()
    # field_selector = "status.successful!=1"
    if team_only:
        operand = "!="
    else:
        operand = "="

    label_selector = f"app=orbit-runner,username{operand}{username}"
    _logger.info("using job selector %s", label_selector)
    try:
        api_response = api_instance.list_namespaced_job(
            namespace=team_name,
            _preload_content=False,
            label_selector=label_selector,
            # field_selector=field_selector,
            watch=False,
        )
        res = json.loads(api_response.data)
    except ApiException as e:
        _logger.info("Exception when calling BatchV1Api->list_namespaced_job: %s\n" % e)
        raise e

    if "items" not in res:
        return []

    return res["items"]


def list_current_pods(label_selector: str = None):
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    load_kube_config()
    api_instance = CoreV1Api()
    try:
        params = dict()
        params["namespace"] = team_name
        params["_preload_content"] = False
        if label_selector:
            params["label_selector"] = label_selector
        api_response = api_instance.list_namespaced_pod(**params)
        res = json.loads(api_response.data)
    except ApiException as e:
        _logger.info("Exception when calling BatchV1Api->list_namespaced_job: %s\n" % e)
        raise e

    if "items" not in res:
        return []

    return res["items"]


def list_storage_pvc():
    load_kube_config()
    api_instance = CoreV1Api()
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    _logger.debug(f"Listing {team_name} namespace persistent volume claims")
    params = dict()
    params["namespace"] = team_name
    params["_preload_content"] = False
    try:
        api_response = api_instance.list_namespaced_persistent_volume_claim(**params)
        res = json.loads(api_response.data)
    except ApiException as e:
        _logger.info("Exception when calling CoreV1Api->list persistent volume claims: %s\n" % e)
        raise e

    if "items" not in res:
        return []

    return res["items"]


def delete_storage_pvc(pvc_name: str):
    load_kube_config()
    api_instance = CoreV1Api()
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    _logger.debug(f"Deleting {team_name} namespace persistent volume claim {pvc_name}")
    params = dict()
    params["name"] = pvc_name
    params["namespace"] = team_name
    params["_preload_content"] = False
    try:
        api_response = api_instance.delete_namespaced_persistent_volume_claim(**params)
        response = {
            "status": str(api_response.status),
            "reason": api_response.reason,
            "message": f"Successfully deleted persistent volume claim={pvc_name}",
        }
    except ApiException as e:
        _logger.info("Exception when calling CoreV1Api->delete persistent volume claim: %s\n" % e)
        e_body = json.loads(e.body)
        response = {"status": str(e_body["code"]), "reason": e_body["reason"], "message": e_body["message"]}

    return response


def list_storage_pv():
    load_kube_config()
    api_instance = CoreV1Api()
    _logger.debug("Listing cluster persistent volumes")
    params = dict()
    params["_preload_content"] = False
    try:
        api_response = api_instance.list_persistent_volume(**params)
        res = json.loads(api_response.data)
    except ApiException as e:
        _logger.info("Exception when calling CoreV1Api->list persistent volumes : %s\n" % e)
        raise e

    if "items" not in res:
        return []

    return res["items"]


def list_storage_class():
    load_kube_config()
    api_instance = StorageV1Api()
    _logger.debug("Listing cluster storage classes")
    params = dict()
    params["_preload_content"] = False
    try:
        api_response = api_instance.list_storage_class(**params)
        res = json.loads(api_response.data)
    except ApiException as e:
        _logger.info("Exception when calling StorageV1Api->list storage class : %s\n" % e)
        raise e

    if "items" not in res:
        return []

    return res["items"]


def delete_job(job_name: str, grace_period_seconds: int = 30):
    props = get_properties()
    global __CURRENT_TEAM_MANIFEST__, __CURRENT_ENV_MANIFEST__
    env_name = props["AWS_ORBIT_ENV"]
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    load_kube_config()
    api_instance = BatchV1Api()
    try:
        api_instance.delete_namespaced_job(
            name=job_name,
            namespace=team_name,
            _preload_content=False,
            grace_period_seconds=grace_period_seconds,
            orphan_dependents=False,
        )
    except ApiException as e:
        _logger.info("Exception when calling BatchV1Api->list_namespaced_job: %s\n" % e)
        raise e


def delete_cronjob(job_name: str, grace_period_seconds: int = 30):
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    load_kube_config()
    api_instance = BatchV1beta1Api()
    try:
        api_instance.delete_namespaced_cron_job(
            name=job_name,
            namespace=team_name,
            _preload_content=False,
            grace_period_seconds=grace_period_seconds,
            orphan_dependents=False,
        )
    except ApiException as e:
        _logger.info("Exception when calling BatchV1Api->delete_namespaced_cron_job: %s\n" % e)
        raise e


def delete_all_my_jobs():
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    username = os.environ.get("JUPYTERHUB_USER", os.environ.get("USERNAME"))
    load_kube_config()
    api_instance = BatchV1Api()
    label_selector = f"app=orbit-runner,username={username}"
    try:
        api_instance.delete_collection_namespaced_job(
            namespace=team_name, _preload_content=False, orphan_dependents=False, label_selector=label_selector
        )
    except ApiException as e:
        _logger.info("Exception when calling BatchV1Api->delete_collection_namespaced_job: %s\n" % e)
        raise e


def list_running_cronjobs():
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    load_kube_config()
    api_instance = BatchV1beta1Api()
    try:
        api_response = api_instance.list_namespaced_cron_job(namespace=team_name, _preload_content=False)
        res = json.loads(api_response.data)
    except ApiException as e:
        _logger.info("Exception when calling BatchV1beta1Api->list_namespaced_cron_job: %s\n" % e)
        raise e

    if "items" not in res:
        return []

    return res["items"]


def _create_eks_job_spec(taskConfiguration: dict, labels: Dict[str, str], team_constants: TeamConstants) -> V1JobSpec:
    """
    Runs Task in Python in a notebook using lambda.

    Parameters
    ----------
    taskConfiguration: dict
        A task definition to execute.

    Returns
    -------
    Response Payload
    """
    props = get_properties()
    global __CURRENT_TEAM_MANIFEST__, __CURRENT_ENV_MANIFEST__
    env_name = props["AWS_ORBIT_ENV"]
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    if __CURRENT_TEAM_MANIFEST__ == None or __CURRENT_TEAM_MANIFEST__["Name"] != team_name:
        __CURRENT_TEAM_MANIFEST__ = load_team_context_from_ssm(env_name, team_name)
    if __CURRENT_ENV_MANIFEST__ == None:
        __CURRENT_ENV_MANIFEST__ = load_env_context_from_ssm(env_name)

    env = build_env(__CURRENT_ENV_MANIFEST__, env_name, taskConfiguration, team_constants, team_name)
    profile = resolve_profile(taskConfiguration, team_constants)
    image = resolve_image(__CURRENT_TEAM_MANIFEST__, profile)
    node_type = get_node_type(taskConfiguration)

    job_name: str = f'run-{taskConfiguration["task_type"]}'
    volumes = team_constants.volumes()
    volume_mounts = team_constants.volume_mounts()
    grant_sudo = False
    if "kubespawner_override" in profile:
        if "volumes" in profile["kubespawner_override"]:
            volumes.extend(profile["kubespawner_override"]["volumes"])
            _logger.info("profile override is attaching volumes: %s", volumes)
        if "volume_mounts" in profile["kubespawner_override"]:
            volume_mounts.extend(profile["kubespawner_override"]["volume_mounts"])
            _logger.info("profile override is mounting volumes: %s", volume_mounts)
    if "compute" in taskConfiguration:
        if "grant_sudo" in taskConfiguration["compute"]:
            if taskConfiguration["compute"]["grant_sudo"] or taskConfiguration["compute"]["grant_sudo"] == "True":
                grant_sudo = True
        if "volumes" in taskConfiguration["compute"]:
            volumes.extend(taskConfiguration["compute"]["volumes"])
            _logger.info("task override is attaching volumes: %s", volumes)
        if "volume_mounts" in taskConfiguration["compute"]:
            volume_mounts.extend(taskConfiguration["compute"]["volume_mounts"])
            _logger.info("task override is mounting volumes: %s", volume_mounts)
        if "labels" in taskConfiguration["compute"]:
            labels = {**labels, **taskConfiguration["compute"]["labels"]}
    node_selector = team_constants.node_selector(node_type)
    _logger.info("volumes:%s", json.dumps(volumes))
    _logger.info("volume_mounts:%s", json.dumps(volume_mounts))
    pod_properties: Dict[str, str] = dict(
        name=job_name,
        image=image,
        cmd=["bash", "-c", "/home/jovyan/.orbit/bootstrap.sh && python /opt/python-utils/notebook_cli.py"],
        port=22,
        image_pull_policy=team_constants.image_pull_policy(),
        image_pull_secrets=None,
        node_selector=node_selector,
        run_as_uid=team_constants.uid(grant_sudo),
        run_as_gid=team_constants.gid(),
        fs_gid=team_constants.gid(),
        run_privileged=False,
        allow_privilege_escalation=True,
        env=env,
        volumes=volumes,
        volume_mounts=volume_mounts,
        labels=labels,
        annotations=team_constants.annotations(),
        lifecycle_hooks=team_constants.life_cycle_hooks(),
        service_account=team_name,
        logger=_logger,
    )
    if grant_sudo:
        pod_properties["uid"] = 0

    if "kubespawner_override" in profile:
        for k, v in profile["kubespawner_override"].items():
            if k in ["image"]:
                # special handling is already done for image
                continue
            if k in pod_properties:
                raise RuntimeError("Override '%s' in profile is not allowed", k)
            _logger.debug("profile overriding pod value %s=%s", k, v)
            pod_properties[k] = v

    pod: V1Pod = make_pod(**pod_properties)
    pod.spec.restart_policy = "Never"
    job_spec = V1JobSpec(
        backoff_limit=0,
        template=pod,
        ttl_seconds_after_finished=120,
    )

    return job_spec


def resolve_image(__CURRENT_TEAM_MANIFEST__, profile):
    if not profile or "kubespawner_override" not in profile or "image" not in profile["kubespawner_override"]:
        repository = __CURRENT_TEAM_MANIFEST__["FinalImageAddress"]
        image = f"{repository}"
    else:
        image = profile["kubespawner_override"]["image"]
    return image


def build_env(__CURRENT_ENV_MANIFEST__, env_name, taskConfiguration, team_constants, team_name):
    env = team_constants.env()
    env["AWS_ORBIT_ENV"] = env_name
    env["AWS_ORBIT_TEAM_SPACE"] = team_name
    env["JUPYTERHUB_USER"] = team_constants.username
    env["USERNAME"] = team_constants.username
    env["AWS_ORBIT_S3_BUCKET"] = __CURRENT_ENV_MANIFEST__["Toolkit"]["S3Bucket"]
    env["task_type"] = taskConfiguration["task_type"]
    env["tasks"] = json.dumps({"tasks": taskConfiguration["tasks"]})
    if "compute" in taskConfiguration:
        env["compute"] = json.dumps({"compute": taskConfiguration["compute"]})
    else:
        env["compute"] = json.dumps({"compute": {"compute_type": "eks", "node_type": "fargate"}})
    return env


def resolve_profile(taskConfiguration, team_constants):
    if "compute" in taskConfiguration and "profile" in taskConfiguration["compute"]:
        profile_name = taskConfiguration["compute"]["profile"]
        profile_name = slugify(profile_name)
        _logger.info(f"using profile %s", profile_name)
        profile = team_constants.profile(profile_name)
        if not profile:
            raise Exception("Profile '%s' not found", profile)
    else:
        profile = team_constants.default_profile()
        _logger.info(f"using default profile %s", profile)
    return profile


def _run_task_eks(taskConfiguration: dict) -> Any:
    """
    Runs Task in Python in a notebook using lambda.

    Parameters
    ----------
    taskConfiguration: dict
        A task definition to execute.

    Returns
    -------
    Response Payload
    """
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    username = os.environ.get("JUPYTERHUB_USER", os.environ.get("USERNAME"))
    node_type = get_node_type(taskConfiguration)
    labels = {"app": f"orbit-runner", "orbit/node-type": node_type, "username": username}
    if node_type == "ec2":
        labels["orbit/attach-security-group"] = "yes"
    team_constants: TeamConstants = TeamConstants()
    job_spec = _create_eks_job_spec(taskConfiguration, labels=labels, team_constants=team_constants)
    load_kube_config()
    if "compute" in taskConfiguration:
        if "labels" in taskConfiguration["compute"]:
            labels = {**labels, **taskConfiguration["compute"]["labels"]}
    job = V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=V1ObjectMeta(
            generate_name=f"orbit-{team_name}-{node_type}-runner-", labels=labels, namespace=team_name
        ),
        spec=job_spec,
    )
    job_instance: V1Job = BatchV1Api().create_namespaced_job(
        namespace=team_name,
        body=job,
    )
    metadata: V1ObjectMeta = job_instance.metadata

    _logger.debug(f"started job {metadata.name}")
    return {
        "ExecutionType": "eks",
        "Identifier": metadata.name,
        "NodeType": node_type,
        "tasks": taskConfiguration["tasks"],
    }


def schedule_python(triggerName: str, frequency: str, taskConfiguration: Dict[str, Any]) -> str:
    """
    Schedule Python Task

    Parameters
    -----------
    triggerName: str
      A unique name of the time trigger that will start this execution.
    frequency: str
       A cron string e.g., cron(0/15 * 1/1 * ? *) to define the starting times of the execution.
    taskConfiguration: dict
         A task definition to execute

        tasks: lst
            A list of python task definition to run.
        module : str
           The python module to run (without .py ext).
        functionName : str
                The python function to start the execution.
        sourcePaths : lst
              A list of s3 python source paths used for importing packages or modules into the application.
        params : dict
             A list of parameters for this task to override the notebook parameters.
        compute : optional, dict
              A list of runtime parameters to control execution.
        container : dict
               A list of parameters to control container execution.
        p_concurrent : str
               The number of parallel threads inside the container that will execute notebooks.
        env_vars : optional, list
              A list of environment parameters to pass to the container.

    Returns
    --------
    arn : str
        The response will be ARN of the event rule created to start this execution:'arn:aws:events:us-east-2:...'

    Example
    --------
    Schedule python task similar to 'run_python' task config. example (refer to 'run_python').
    """
    taskConfiguration["task_type"] = "python"
    if "compute_type" not in taskConfiguration or taskConfiguration["compute_type"] == "eks":
        return schedule_task_eks(triggerName, frequency, taskConfiguration)
    else:
        raise RuntimeError("Unsupported compute_type '%s'", taskConfiguration["compute_type"])


def schedule_notebooks(triggerName: str, frequency: str, taskConfiguration: dict) -> str:
    """
    Schedules Notebooks
    Parameters
    ---------
    triggerName: str
        A unique name of the time trigger that will start this exec
    frequency: str
        A cron string e.g., cron(0/15 * 1/1 * ? *) to define the starting times of the execution
    taskConfiguration: dict
        A task definition to execute

        notebooks : lst
            A list of notebook task definition to run.
        notebookName : str
            The filename of the notebook.
        notebookName : str
            The relative path to the notebook file starting at the repository root.
        targetPath : str
            The target S3 directory where the output notebook and all related output will be generated.
        params : dict
            A list of parameters for this task to override the notebook parameters.
        compute : optional, dict
            A list of runtime parameters to control execution.
        container : dict
            A list of parameters to control container execution.
        p_concurrent : str
            The number of parallel threads inside the container that will execute notebooks.
        env_vars : optional, list
            A list of environment parameters to pass to the container.

    Returns
    -------
    arn: str
        The response will be ARN of the event rule created to start this execution: 'arn:aws:events:us-east-2:...'

    Example
    --------
    Schedule python task similar to 'run_notebook' task config. example (refer to 'run_notebook').
    """
    taskConfiguration["task_type"] = "jupyter"
    if "compute_type" not in taskConfiguration or taskConfiguration["compute_type"] == "eks":
        return schedule_task_eks(triggerName, frequency, taskConfiguration)
    else:
        raise RuntimeError("Unsupported compute_type '%s'", taskConfiguration["compute_type"])


def get_node_type(taskConfiguration) -> str:
    if "compute" not in taskConfiguration or "node_type" not in taskConfiguration["compute"]:
        node_type = "fargate"
    else:
        node_type = taskConfiguration["compute"]["node_type"]

    return node_type


def schedule_task_eks(triggerName: str, frequency: str, taskConfiguration: dict) -> Any:
    """
    Parameters
    ----------
    triggerName: str
        A unique name of the time trigger that will start this exec
    frequency: str
        A cron string e.g., cron(0/15 * 1/1 * ? *) to define the starting times of the execution
    taskConfiguration: Any
    Return
    ------
    Example
    -------
    Example for schedule_task similar to run_python and run_notebook tasks (refer to 'run_python').
    """
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    node_type = get_node_type(taskConfiguration)
    username = os.environ.get("JUPYTERHUB_USER", os.environ.get("USERNAME"))
    cronjob_id = f"orbit-{team_name}-{triggerName}"
    labels = {"app": f"orbit-runner", "orbit/node-type": node_type, "username": username, "cronjob_id": cronjob_id}
    team_constants: TeamConstants = TeamConstants(username)
    job_spec = _create_eks_job_spec(taskConfiguration, labels=labels, team_constants=team_constants)
    cron_job_template: V1beta1JobTemplateSpec = V1beta1JobTemplateSpec(spec=job_spec)
    cron_job_spec: V1beta1CronJobSpec = V1beta1CronJobSpec(job_template=cron_job_template, schedule=frequency)
    job = V1beta1CronJob(
        api_version="batch/v1beta1",
        kind="CronJob",
        metadata=V1ObjectMeta(name=cronjob_id, labels=labels, namespace=team_name),
        status=V1beta1CronJobStatus(),
        spec=cron_job_spec,
    )
    load_kube_config()

    job_instance: V1beta1CronJob = BatchV1beta1Api().create_namespaced_cron_job(namespace=team_name, body=job)
    metadata: V1ObjectMeta = job_instance.metadata
    return {
        "ExecutionType": "eks",
        "Identifier": metadata.name,
    }
    metadata: V1ObjectMeta = job_instance.metadata
    _logger.debug(f"started job {metadata.name}")


def load_kube_config():
    if "AWS_WEB_IDENTITY_TOKEN_FILE" in os.environ and "eks.amazonaws.com" in os.environ["AWS_WEB_IDENTITY_TOKEN_FILE"]:
        k8_config.load_incluster_config()
    else:
        k8_config.load_kube_config()


def delete_task_schedule(triggerName: str, compute_type: str = "eks") -> None:
    if compute_type == "eks":
        return delete_task_schedule_eks(triggerName)
    else:
        raise RuntimeError("Unsupported compute_type '%s'", compute_type)


def delete_task_schedule_eks(triggerName: str) -> None:
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]
    load_kube_config()

    BatchV1beta1Api().delete_namespaced_cron_job(name=f"orbit-{team_name}-{triggerName}", namespace=team_name)


def order_def(task_definition: Any) -> datetime:
    """
    Order task definitions by their timestamps of when they start at.

    Parameters
    -----------
    task_definition: Any

    Returns
    --------
    date : datetime
        Timestamp for a specific task definition, or the current timestamp if one does not already exist.
    """

    if "epoch" in task_definition:
        return task_definition["epoch"]
    return datetime.now().timestamp()


def wait_for_tasks_to_complete(
    tasks: List[Dict[str, str]],
    delay: Optional[int] = 10,
    maxAttempts: Optional[int] = 10,
    tail_log: Optional[bool] = False,
) -> bool:
    """
    Parameters
    ----------
    tasks: lst
       A list of structures with container type, execution arn, and job arn.
    delay: int
       Number of seconds to wait until checking containers state again (default = 60).
    maxAttempts: str
        Number of attempts to check if containers stopped before returning a failure (default = 10).
    tail_log: bool
       if True, will tail the log of the containers until they are stopped (default = false).

    Returns
    -------
    None
        None.

    Example
    --------
    >>> from aws_orbit_sdk.controller import wait_for_tasks_to_complete
    controller.wait_for_tasks_to_complete(containers, 60,40)
    """

    completed_tasks = []
    errored_tasks = []
    attempts = 0
    props = get_properties()
    team_name = props["AWS_ORBIT_TEAM_SPACE"]

    incomplete_tasks = []
    _logger.info("Waiting2 for %s tasks %s", len(tasks), tasks)
    load_kube_config()
    while True:
        for task in tasks:
            _logger.debug("Checking execution state of: %s", task)
            try:
                current_jobs: V1JobList = BatchV1Api().list_namespaced_job(
                    namespace=team_name, label_selector=f"app=orbit-runner"
                )
            except exceptions.ApiException as e:
                _logger.error("Error during list jobs for %s: %s", team_name, e)
                # try again after 5 seconds.
                time.sleep(5)
                current_jobs: V1JobList = BatchV1Api().list_namespaced_job(
                    namespace=team_name, label_selector=f"app=orbit-runner"
                )
            task_name = task["Identifier"]
            for job in current_jobs.items:
                job_instance: V1Job = cast(V1Job, job)
                job_metadata: V1ObjectMeta = cast(V1ObjectMeta, job_instance.metadata)
                if job_metadata.name != task["Identifier"]:
                    continue
                job_status: V1JobStatus = cast(V1JobStatus, job_instance.status)
                _logger.debug(f"job-status={job_status}")
                if job_status.active == 1:
                    incomplete_tasks.append(task)
                    _logger.info("Task %s is running with status %s", task, job_status)
                elif job_status.failed == 1:
                    _logger.debug(f"Execution error: {task_name}")
                    errored_tasks.append(task)
                elif job_status.succeeded:
                    completed_tasks.append(task)
                else:
                    incomplete_tasks.append(task)

            if tail_log:
                tail_logs(team_name, tasks)

        tasks = incomplete_tasks
        incomplete_tasks = []
        _logger.info(f"Running: {len(tasks)} Completed: {len(completed_tasks)} Errored: {len(errored_tasks)}")

        attempts += 1
        if len(tasks) == 0:
            _logger.info("All tasks stopped")
            return len(errored_tasks) == 0
        elif attempts >= maxAttempts:
            _logger.info("Stopped waiting as maxAttempts reached")
            return len(errored_tasks) > 0
        else:
            _logger.info("waiting for %s", tasks)
            time.sleep(delay)
            load_kube_config()


def tail_logs(team_name, tasks) -> None:
    for task in tasks:
        task_id = task["Identifier"]
        _logger.info("Watching task: '%s'", task_id)

        current_pods: V1PodList = CoreV1Api().list_namespaced_pod(
            namespace=team_name, label_selector=f"job-name={task_id}"
        )
        for pod in current_pods.items:
            pod_instance: V1Pod = cast(V1Pod, pod)
            _logger.debug("pod: %s", pod_instance.metadata.name)
            pod_status: V1PodStatus = cast(V1PodStatus, pod_instance.status)
            _logger.debug("pod s: %s", pod_status)
            if pod_status.conditions:
                for c in pod_status.conditions:
                    condition: V1PodCondition = cast(V1PodCondition, c)
                    if condition.type == "Failed" or condition.reason == "Unschedulable":
                        _logger.info("pod has error status %s , %s", condition.reason, condition.message)
                        return
            if pod_status.container_statuses:
                for s in pod_status.container_statuses:
                    container_status: V1ContainerStatus = cast(V1ContainerStatus, s)
                    container_state: V1ContainerState = container_status.state
                    _logger.debug("task status: %s ", container_status)
                    if container_status.started or container_state.running or container_state.terminated:
                        _logger.info("task %s status: %s", pod_instance.metadata.name, container_state)
                        w = k8_watch.Watch()
                        for line in w.stream(
                            CoreV1Api().read_namespaced_pod_log, name=pod_instance.metadata.name, namespace=team_name
                        ):
                            _logger.info(line)
                    else:
                        _logger.info("task not started yet for %s", task_id)


def logEvents(paginator: Any, logGroupName: Any, logStreams: Any, fromTime: Any) -> int:
    """
    Parameters
    ----------
    paginator: Any
    logGroupName: Any
    logStreams: Any
    fromTime: Any

    Return
    -------

    Example
    -------
    """
    logging.basicConfig(format="%(message)s", level=logging.INFO, datefmt="%Y-%m-%d %H:%M:%S")
    event_logger = logging.getLogger("event")

    response_iterator = paginator.paginate(
        logGroupName=logGroupName,
        logStreamNames=logStreams,
        interleaved=True,
        startTime=fromTime,
        PaginationConfig={"MaxItems": 10, "PageSize": 10},
    )
    lastTime = 0
    while True:
        for page in response_iterator:
            for e in page["events"]:
                t = datetime.fromtimestamp(e["timestamp"] / 1000).strftime("%Y-%m-%d %H:%M:%S")
                m = e["message"]
                _logger.info(t, m)
                lastTime = e["timestamp"]

            if "nextToken" in page.keys():
                marker = page["nextToken"]
            else:
                marker = None

        if marker == None:
            return lastTime + 1
        else:
            response_iterator = paginator.paginate(
                logGroupName=logGroupName,
                logStreamNames=logStreams,
                interleaved=True,
                PaginationConfig={
                    "MaxItems": 500,
                    "PageSize": 100,
                    "StartingToken": marker,
                },
            )


def all_tasks_stopped(tasks_state: Any) -> bool:
    """
    Checks if all tasks are stopped or if any are still running.

    Parameters
    ---------
    tasks_state: Any
        Task state dictionary object

    Returns
    --------
    response:  bool
        True if all tasks are stopped.

    """
    for t in tasks_state["tasks"]:
        if t["lastStatus"] in ("PENDING", "RUNNING"):
            return False
    return True
